import random
import re
from typing import Any, Callable, Dict, Set, Type, TypeVar

from pydantic import BaseModel, ValidationError

from .error import AutumnHTTPError, AutumnValidationError

T = TypeVar("T", bound=BaseModel)


_CONV_RE = re.compile(r"_([a-z])")


def _snake_to_camel(snake_str: str) -> str:
    return _CONV_RE.sub(lambda match: match.group(1).upper(), snake_str)


def _decompose_value(value: Any) -> Any:
    if isinstance(value, BaseModel):
        return value.model_dump()
    elif isinstance(value, (list, tuple, set)):
        return [_decompose_value(item) for item in value]

    return value


def _build_payload(
    scope: Dict[str, Any], method: Callable, *, ignore: Set[str] = set()
) -> Dict[str, Any]:
    params = method.__code__.co_varnames

    camel_to_snake_param_map = {}
    for p in params:
        camel_case = _snake_to_camel(p)
        if camel_case != p:
            camel_to_snake_param_map[camel_case] = p

    payload: Dict[str, Any] = {}

    for key, value in scope.items():
        payload_param = key
        if payload_param in camel_to_snake_param_map:
            payload_param = camel_to_snake_param_map[payload_param]

        if (
            payload_param != "self"
            and (payload_param in params)
            and (payload_param not in ignore)
            and (value is not None)
        ):
            payload[payload_param] = _decompose_value(value)

    return payload


def _build_model(model: Type[T], data: Dict[str, Any]) -> T:
    try:
        return model.model_validate(data)
    except ValidationError as e:
        errors = e.errors()
        error_message = errors[0]["msg"]
        error_path = errors[0]["loc"]
        error_code = errors[0]["type"]

        err = AutumnValidationError(
            f"{error_message} at {error_path} with code {error_code}",
            "validation_error",
        )
        err.add_note("Received response: " + str(data))
        raise err


def _check_response(status_code: int, data: Dict[str, Any]) -> None:
    if not 200 <= status_code < 300:
        message = data.get("message", "No error message provided.")
        code = data.get("code", "unknown_error")
        raise AutumnHTTPError(
            message,
            code,
            status_code,
        )


class ExponentialBackoff:
    def __init__(self):
        rand = random.Random()
        rand.seed()

        self._rand = rand
        self._base = 2
        self._state = 0

    def tick(self):
        self._state += 1

    @property
    def bedtime(self):
        raw_time = self._base**self._state
        jitter = self._rand.uniform(0, 1)
        return raw_time + jitter



from __future__ import annotations

from typing import TYPE_CHECKING

from . import _build_response, _extract

if TYPE_CHECKING:
    from starlette.requests import Request


async def list_products_route(request: Request):
    identify, autumn, _ = await _extract(request, get_json=False)

    customer_id = identify["customer_id"]

    response = await autumn.products.list(customer_id=customer_id)
    return _build_response(response)



import pytest

from autumn.utils import _build_payload as _build_kwargs


def _temp_method(customer_id: str, feature_id: str, *, range: str):
    pass

def test_build_kwargs():
    json_data = {
        "open_in_new_tab": True,
        "openInNewTab": True,
        "customer_id": "cus_id",
        "feature_id": "feature_id",
        "range": "24h",
    }

    kwargs = _build_kwargs(json_data, _temp_method)
    
    _temp_method(**kwargs)


from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union

from .customers import Customers
from .entities import Entities
from .features import Features
from .http import HTTPClient
from .models.response import (
    AttachResponse,
    CancelResponse,
    CheckoutResponse,
    CheckResponse,
    QueryResponse,
    TrackResponse,
)
from .products import Products
from .utils import _build_payload

if TYPE_CHECKING:
    from .models.features import Feature
    from .models.meta import CustomerData, FeatureOptions, ProductOptions

__all__ = ("Client", )


class Client:
    """
    The main client class for interacting with the Autumn API.

    Example:

    .. code-block:: python

        import autumn

        client = autumn.Client(token="your_api_key")

        # Attach a customer to a product
        client.attach(
            customer_id="john_doe",
            product_id="chat_messages",
        )

    .. note::
        This client should not be used in async contexts. Use :class:`~autumn.aio.client.AsyncClient` instead.
        The :class:`~autumn.aio.client.AsyncClient` class is a *wrapper* around the :class:`~autumn.Client` class that provides async support.
        It works the same, but you must ``await`` your method calls.

        .. code-block:: python

            import asyncio

            from autumn.aio import (
                Client,
            )


            async def main():
                client = Client(token="your_api_key")
                await client.attach(
                    customer_id="john_doe",
                    product_id="chat_messages",
                )


            asyncio.run(main())

        The async client requires ``aiohttp`` to be installed. You can install it via: ``pip install aiohttp``.

    Parameters
    ----------
    token: str
        The API key to use for authentication.
    max_retries: int
        The maximum number of retries to attempt for failed requests.
    base_url: Optional[str]
        The base URL of the Autumn API. This is useful when you are self-hosting Autumn and need to point to your own instance.

    Attributes
    ----------
    customers: :class:`~autumn.customers.Customers`
        An interface to Autumn's customer API.
    features: :class:`~autumn.features.Features`
        An interface to Autumn's feature API.
    products: :class:`~autumn.products.Products`
        An interface to Autumn's product API.
    entities: :class:`~autumn.entities.Entities`
        An interface to Autumn's entities API.
    """

    def __init__(
        self,
        token: str,
        version: str = "1.2",
        *,
        base_url: Optional[str] = None,
        max_retries: int = 5,
    ):
        from . import BASE_URL

        _base_url = base_url or BASE_URL
        _base_url = _base_url.rstrip("/")

        attempts = max_retries + 1  # account for the original request
        self.http = HTTPClient(_base_url, version, token, attempts=attempts)
        self.customers = Customers(self.http)
        self.features = Features(self.http)
        self.products = Products(self.http)
        self.entities = Entities(self.http)

    def checkout(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        products: Optional[List[ProductOptions]] = None,
        success_url: Optional[str] = None,
        force_checkout: bool = False,
        options: Optional[List[FeatureOptions]] = None,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
        checkout_session_params: Optional[Dict[str, Any]] = None,
        reward: Optional[Union[str, List[str]]] = None,
    ) -> CheckoutResponse:
        """Checkout a customer for a product.

        Parameters
        ----------
        customer_id: str
            The ID of the customer to checkout.
        product_id: Optional[str]
            The ID of the product to checkout.
        products: Optional[List[ProductOptions]]
            The products to checkout.
        success_url: Optional[str]
            The URL to redirect to after a successful checkout.
        force_checkout: bool
            Whether to force the customer to checkout.
        entity_id: Optional[str]
            The ID of the entity to checkout.
        customer_data: Optional[CustomerData]
            The customer data to checkout.
        reward: Optional[str | List[str]]
            The reward to checkout. Can pass in an array too.
        """

        payload = _build_payload(locals(), self.checkout)
        return self.http.request("POST",
                                 "/checkout",
                                 CheckoutResponse,
                                 json=payload)

    def attach(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        product_ids: Optional[List[str]] = None,
        products: Optional[List[ProductOptions]] = None,
        success_url: Optional[str] = None,
        force_checkout: bool = False,
        features: Optional[List[Feature]] = None,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
        free_trial: Optional[bool] = None,
        options: Optional[List[FeatureOptions]] = None,
        reward: Optional[str | List[str]] = None,
    ) -> AttachResponse:
        """Attach a customer to a product.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to attach.
        product_id: Optional[str]
            The ID of the product to attach.
        product_ids: Optional[List[str]]
            The IDs of the products to attach.
        products: Optional[List[AttachProductOptions]]
            The products to attach.
        success_url: Optional[str]
            The URL to redirect to after a successful attachment.
        force_checkout: bool
            Whether to force the customer to checkout.
        features: Optional[List[Feature]]
            The features to attach.
        entity_id: Optional[str]
            The ID of the entity to attach.
        customer_data: Optional[CustomerData]
            The customer data to attach.
        free_trial: Optional[bool]
            Whether to attach a free trial.
        options: Optional[List[FeatureOptions]]
            The options to attach.
        reward: Optional[str | List[str]]
            The reward to attach. Can pass in an array too.
        Returns
        -------
        :class:`~autumn.models.response.AttachResponse`
            The response from the API.
        """

        assert (product_id is not None or product_ids is not None
                ), "Either product_id or product_ids must be provided"
        assert not (product_id is not None and product_ids is not None
                    ), "Only one of product_id or product_ids must be provided"

        payload = _build_payload(locals(), self.attach)
        return self.http.request("POST",
                                 "/attach",
                                 AttachResponse,
                                 json=payload)

    def check(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        feature_id: Optional[str] = None,
        required_balance: Optional[int] = 1,
        send_event: bool = False,
        with_preview: bool = False,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
    ) -> CheckResponse:
        """Check if a customer has access to a product or feature.

        |maybecoro|

        You must pass either ``product_id`` or ``feature_id``. Failure to pass one and **only one** will raise an assertion error.

        Parameters
        ----------
        customer_id: str
            The ID of the customer to check.
        product_id: Optional[str]
            The ID of the product to check.
        feature_id: Optional[str]
            The ID of the feature to check.
        required_balance: Optional[int]
            The required balance to check.
        send_event: bool
            Whether to record a usage event with checking access. The ```required_balance`` field will be used as the usage ``value``.
        with_preview: bool
            If true, the response will include a ``preview`` object, which can be used to display information such as a paywall or upgrade confirmation.
        entity_id: Optional[str]
            If using entity balances (eg, seats), the entity ID to check access for.
        customer_data: Optional[CustomerData]
            Additional customer properties. These will be used if the customer's properties are not already set.

        Returns
        -------
        :class:`~autumn.models.response.CheckResponse`
            The response from the API.
        """

        assert (product_id is not None or feature_id is not None
                ), "Either product_id or feature_id must be provided"

        payload = _build_payload(locals(), self.check)
        return self.http.request("POST", "/check", CheckResponse, json=payload)

    def track(
        self,
        customer_id: str,
        feature_id: Optional[str] = None,
        *,
        value: int = 1,
        entity_id: Optional[str] = None,
        event_name: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
        customer_data: Optional[CustomerData] = None,
    ) -> TrackResponse:
        """
        Track a feature usage.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to track.
        feature_id: Optional[str]
            The ID of the feature to track. This or the ``event_name`` must be provided.
        value: int
            The amount of the feature to deduct.
        entity_id: Optional[str]
            The ID of the entity to track.
        event_name: Optional[str]
            The name of the event to track.
        idempotency_key: Optional[str]
            A unique identifier for the track event. If not provided, the SDK will not generate one - the Autumn API does not expect one.
        properties: Optional[Dict[str, Any]]
            Additional properties to track.
        customer_data: Optional[CustomerData]
            Additional customer properties. These will be used if the customer's properties are not already set.

        Returns
        -------
        :class:`~autumn.models.response.TrackResponse`
            The response from the API.
        """
        assert (feature_id or
                event_name), "Either feature_id or event_name must be provided"
        payload = _build_payload(locals(), self.track)
        return self.http.request("POST", "/track", TrackResponse, json=payload)

    def query(
        self,
        customer_id: str,
        feature_id: Union[str, List[str]],
        *,
        range: Literal["24h", "7d", "30d", "90d", "last_cycle"] = "30d",
    ) -> QueryResponse:
        """
        Query usage analytics for a customer on a specific feature.

        Parameters
        ----------
        customer_id: str
            The ID of the customer to query for.
        feature_id:
            The ID of the feature you want to query analytics for.
        range: Literal["24h", "7d", "30d", "90d", "last_cycle"]
            Analytics time period.

        Returns
        -------
        :class:`~autumn.models.response.QueryResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.query)
        return self.http.request("POST", "/query", QueryResponse, json=payload)

    def cancel(
        self,
        customer_id: str,
        product_id: str,
        *,
        entity_id: Optional[str] = None,
        cancel_immediately: bool = False,
    ) -> CancelResponse:
        """Cancel a product for a customer.

        Parameters
        ----------
        customer_id: str
            The ID of the customer to cancel the product for.
        product_id: str
            The ID of the product to cancel.
        entity_id: Optional[str]
            The ID of the entity to cancel the product for.
        cancel_immediately: bool
            Whether to cancel the product immediately. If false, the product will be cancelled at the end of the billing cycle.

        Returns
        -------
        :class:`~autumn.models.response.CancelResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.cancel)
        return self.http.request(
            "POST",
            "/cancel",
            CancelResponse,
            json=payload,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from . import _build_response, _extract

if TYPE_CHECKING:
    from starlette.requests import Request


async def create_customer_route(request: Request):
    identify, autumn, _ = await _extract(request)

    customer_id = identify["customer_id"]
    customer_data = identify["customer_data"]
    response = await autumn.customers.create(customer_id, **customer_data)
    return _build_response(response)


async def pricing_table_route(request: Request):
    identify, autumn, _ = await _extract(request, get_json=False)

    customer_id = identify["customer_id"]
    response = await autumn.customers.pricing_table(customer_id)
    return _build_response(response)



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .env import AppEnv
from .products import ProductItemInterval

__all__ = (
    "ProductStatus",
    "FeatureType",
    "CustomerInvoice",
    "CustomerFeature",
    "CustomerProduct",
    "Customer",
    "PriceInfo",
    "ItemInfo",
    "GetPricingTableParams",
    "PricingTableProduct",
)


class ProductStatus(str, Enum):
    ACTIVE = "active"
    EXPIRED = "expired"
    TRIALING = "trialing"
    SCHEDULED = "scheduled"


class FeatureType(str, Enum):
    STATIC = "static"
    CONTINUOUS_USE = "continuous_use"
    SINGLE_USE = "single_use"
    BOOLEAN = "boolean"


class CustomerInvoice(BaseModel):
    product_ids: List[str]
    stripe_id: str
    status: str
    total: float
    currency: str
    created_at: int
    hosted_invoice_url: str


class CustomerFeature(BaseModel):
    id: str
    name: str
    unlimited: Optional[bool] = None
    type: Optional[FeatureType] = None
    interval: Optional[ProductItemInterval] = None
    balance: Optional[float] = None
    usage: Optional[float] = None
    included_usage: Optional[float] = None
    next_reset_at: Optional[int] = None
    breakdown: Optional[List[Dict[str, Any]]] = None


class ReferredCustomer(BaseModel):
    id: str
    name: Optional[str] = None
    email: Optional[str] = None


class CustomerReferral(BaseModel):
    program_id: str
    customer: ReferredCustomer
    reward_applied: bool
    created_at: int


class CustomerProduct(BaseModel):
    id: str
    name: Optional[str] = None
    group: Optional[str] = None
    status: ProductStatus
    started_at: int
    canceled_at: Optional[int] = None
    subscription_ids: Optional[List[str]] = None
    current_period_start: Optional[int] = None
    current_period_end: Optional[int] = None
    quantity: Optional[int] = None


class Customer(BaseModel):
    id: Optional[str] = None
    created_at: int
    name: Optional[str] = None
    email: Optional[str] = None
    fingerprint: Optional[str] = None
    stripe_id: Optional[str] = None
    env: AppEnv
    metadata: Dict[str, Any]
    products: List[CustomerProduct]
    features: Dict[str, CustomerFeature]
    invoices: Optional[List[CustomerInvoice]] = None
    payment_method: Optional[Any] = None
    referrals: Optional[List[CustomerReferral]] = None


class PriceInfo(BaseModel):
    primaryText: str
    secondaryText: Optional[str] = None


class ItemInfo(BaseModel):
    primaryText: str
    secondaryText: Optional[str] = None


class GetPricingTableParams(BaseModel):
    customer_id: Optional[str] = None


class PricingTableProduct(BaseModel):
    id: str
    name: str
    buttonText: str
    price: PriceInfo
    items: List[ItemInfo]



from unittest.mock import AsyncMock

import pytest
from pydantic import BaseModel
from starlette.responses import JSONResponse
from typing import Optional

from autumn.asgi.routes import core
from autumn.client import Client
from autumn.models.response import AttachResponse, CheckResponse, TrackResponse


DUMMY_CUSTOMER_ID = "user_123"
DUMMY_AUTUMN_SECRET_KEY = "sk_test_a1b2c3d4e5f6g7h8i9j0"
DUMMY_IDENTIFY_RETURN_VALUE = {"customer_id": DUMMY_CUSTOMER_ID, "customer_data": {}}


class DummyRequest:
    def __init__(self, state_obj):
        self.state = state_obj
        self._json = None

    async def json(self):
        return getattr(self, "_json", None)


def _setup_http_mock(autumn: Client, mock_response: Optional[BaseModel]) -> None:
    autumn.http.request = AsyncMock(return_value=mock_response)  # type: ignore


def _prepare_request(
    json_payload=None, mock_response: Optional[BaseModel] = None
) -> DummyRequest:
    autumn = Client(token=DUMMY_AUTUMN_SECRET_KEY)

    _setup_http_mock(autumn, mock_response)

    identify = AsyncMock(return_value=DUMMY_IDENTIFY_RETURN_VALUE)
    autumn_obj = {"identify": identify, "client": autumn}

    State = type("State", (), {})
    state_obj = State()
    setattr(state_obj, "__autumn__", autumn_obj)

    request = DummyRequest(state_obj)
    if json_payload is not None:
        request._json = json_payload

    return request


@pytest.mark.asyncio
async def test_attach_route_with_product_id():
    json_payload = {"product_id": "product_123"}

    mock_response = AttachResponse(
        code="success",
        message="Attached",
        checkout_url="https://checkout.example.com",
        customer_id=DUMMY_CUSTOMER_ID,
        product_ids=["product_123"],
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.attach_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_attach_route_with_product_id_camel_case():
    json_payload = {"productId": "product_123"}

    mock_response = AttachResponse(
        code="success",
        message="Attached",
        checkout_url="https://checkout.example.com",
        customer_id=DUMMY_CUSTOMER_ID,
        product_ids=["product_123"],
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.attach_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_attach_route_with_product_ids():
    json_payload = {"product_ids": ["product_123", "product_456"]}

    mock_response = AttachResponse(
        code="success",
        message="Attached",
        checkout_url="https://checkout.example.com",
        customer_id=DUMMY_CUSTOMER_ID,
        product_ids=["product_123", "product_456"],
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.attach_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_attach_route_with_product_ids_camel_case():
    json_payload = {"productIds": ["product_123", "product_456"]}

    mock_response = AttachResponse(
        code="success",
        message="Attached",
        checkout_url="https://checkout.example.com",
        customer_id=DUMMY_CUSTOMER_ID,
        product_ids=["product_123", "product_456"],
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.attach_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_attach_route_without_product_id_or_product_ids():
    json_payload = {}

    request = _prepare_request(json_payload)

    try:
        await core.attach_route(request)  # type: ignore
        assert False, "Should have raised AssertionError"
    except AssertionError as e:
        assert "Either product_id or product_ids must be provided" in str(e)


@pytest.mark.asyncio
async def test_check_route_with_product_id():
    json_payload = {"product_id": "product_123"}

    mock_response = CheckResponse(
        allowed=True,
        code="success",
        balance=100.0,
        customer_id=DUMMY_CUSTOMER_ID,
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.check_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_check_route_with_product_id_camel_case():
    json_payload = {"productId": "product_123"}

    mock_response = CheckResponse(
        allowed=True,
        code="success",
        balance=100.0,
        customer_id=DUMMY_CUSTOMER_ID,
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.check_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_check_route_with_feature_id():
    json_payload = {"feature_id": "feature_123"}

    mock_response = CheckResponse(
        allowed=True,
        code="success",
        balance=100.0,
        customer_id=DUMMY_CUSTOMER_ID,
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.check_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_check_route_with_feature_id_camel_case():
    json_payload = {"featureId": "feature_123"}

    mock_response = CheckResponse(
        allowed=True,
        code="success",
        balance=100.0,
        customer_id=DUMMY_CUSTOMER_ID,
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.check_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_check_route_without_product_id_or_feature_id():
    json_payload = {}

    request = _prepare_request(json_payload)

    try:
        await core.check_route(request)  # type: ignore
        assert False, "Should have raised AssertionError"
    except AssertionError as e:
        assert "Either product_id or feature_id must be provided" in str(e)


@pytest.mark.asyncio
async def test_track_route_with_feature_id():
    json_payload = {"feature_id": "feature_123"}

    mock_response = TrackResponse(
        id="track_123",
        code="success",
        customer_id=DUMMY_CUSTOMER_ID,
        feature_id="feature_123",
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.track_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_track_route_with_feature_id_camel_case():
    json_payload = {"featureId": "feature_123"}

    mock_response = TrackResponse(
        id="track_123",
        code="success",
        customer_id=DUMMY_CUSTOMER_ID,
        feature_id="feature_123",
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.track_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_track_route_with_event_name():
    json_payload = {"event_name": "event_123"}

    mock_response = TrackResponse(
        id="track_123",
        code="success",
        customer_id=DUMMY_CUSTOMER_ID,
        event_name="event_123",
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.track_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_track_route_with_event_name_camel_case():
    json_payload = {"eventName": "event_123"}

    mock_response = TrackResponse(
        id="track_123",
        code="success",
        customer_id=DUMMY_CUSTOMER_ID,
        event_name="event_123",
    )

    request = _prepare_request(json_payload, mock_response)

    response = await core.track_route(request)  # type: ignore

    assert isinstance(response, JSONResponse)


@pytest.mark.asyncio
async def test_track_route_without_feature_id_or_event_name():
    json_payload = {}

    request = _prepare_request(json_payload)

    try:
        await core.track_route(request)  # type: ignore
        assert False, "Should have raised AssertionError"
    except AssertionError as e:
        assert "Either feature_id or event_name must be provided" in str(e)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Generic,
    List,
    TypeVar,
    Union,
    overload,
)

from .models.balance import Balance
from .models.meta import Empty
from .utils import _build_payload

if TYPE_CHECKING:
    from .aio.http import AsyncHTTPClient
    from .http import HTTPClient

T_HttpClient = TypeVar("T_HttpClient", "AsyncHTTPClient", "HTTPClient")


class Features(Generic[T_HttpClient]):
    """An interface to Autumn's feature API.

    .. warning::
        This class is not intended for public use. It is used internally by the :class:`autumn.Client` class.
        Do not initialize this class directly.
    """

    def __init__(self, http: T_HttpClient):
        self._http = http

    @overload
    def set_usage(
        self: "Features[HTTPClient]",
        customer_id: str,
        feature_id: str,
        value: int,
    ) -> Empty: ...

    @overload
    def set_usage(
        self: "Features[AsyncHTTPClient]",
        customer_id: str,
        feature_id: str,
        value: int,
    ) -> Coroutine[Any, Any, Empty]: ...

    def set_usage(
        self, customer_id: str, feature_id: str, value: int
    ) -> Union[Empty, Coroutine[Any, Any, Empty]]:
        """Set the usage of a feature for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to set the usage for.
        feature_id: str
            The ID of the feature to set the usage for.
        value: int
            The amount the usage should be updated to.

        Returns
        -------
        :class:`~autumn.models.response.Empty`
            This is a placeholder type. Treat it as :class:`None`.
        """
        payload = _build_payload(
            locals(), Features.set_usage, ignore={"customer_id"}
        )
        return self._http.request(
            "POST",
            f"/customers/{customer_id}/balances",
            Empty,
            json=payload,
        )

    @overload
    def set_balances(
        self: "Features[HTTPClient]",
        customer_id: str,
        balances: List[Balance],
    ) -> Empty: ...

    @overload
    def set_balances(
        self: "Features[AsyncHTTPClient]",
        customer_id: str,
        balances: List[Balance],
    ) -> Coroutine[Any, Any, Empty]: ...

    def set_balances(
        self, customer_id: str, balances: List[Balance]
    ) -> Union[Empty, Coroutine[Any, Any, Empty]]:
        """Set the balances of a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to set the balances for.
        balances: List[:class:`~autumn.models.balance.Balance`]
            The balances to set for the customer.

        Returns
        -------
        :class:`~autumn.models.response.Empty`
            This is a placeholder type. Treat it as :class:`None`.
        """

        payload = _build_payload(
            locals(), Features.set_balances, ignore={"customer_id"}
        )
        return self._http.request(
            "POST", f"/customers/{customer_id}/balances", Empty, json=payload
        )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Generic,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from .models.entities import Entity
from .models.meta import Empty
from .models.response import TransferProductResponse
from .utils import _build_payload

if TYPE_CHECKING:
    from .aio.http import AsyncHTTPClient
    from .http import HTTPClient

T_HttpClient = TypeVar("T_HttpClient", "AsyncHTTPClient", "HTTPClient")

__all__ = ("Entities",)


class Entities(Generic[T_HttpClient]):
    """An interface to Autumn's entities API.

    .. warning::
        This class is not intended for public use. It is used internally by the :class:`autumn.Client` class.
        Do not initialize this class directly.
    """

    def __init__(self, http: T_HttpClient):
        self._http = http

    @overload
    def get(
        self: "Entities[HTTPClient]",
        customer_id: str,
        entity_id: str,
        expand: Optional[List[Literal["invoices"]]] = None,
    ) -> Entity: ...

    @overload
    def get(
        self: "Entities[AsyncHTTPClient]",
        customer_id: str,
        entity_id: str,
        expand: Optional[List[Literal["invoices"]]] = None,
    ) -> Coroutine[Any, Any, Entity]: ...

    def get(
        self,
        customer_id: str,
        entity_id: str,
        expand: Optional[List[Literal["invoices"]]] = None,
    ) -> Union[Entity, Coroutine[Any, Any, Entity]]:
        """Get an entity by their ID.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get.
        entity_id: str
            The ID of the entity to get.
        expand: Optional[List[Literal["invoices"]]]
            Additional fields to expand in the response.

        Returns
        -------
        :class:`~autumn.models.entities.Entity`
            The entity.
        """
        params = {}
        if expand is not None:
            params["expand"] = expand
        return self._http.request(
            "GET",
            f"/customers/{customer_id}/entities/{entity_id}",
            Entity,
            params=params,
        )

    @overload
    def create(
        self: "Entities[HTTPClient]",
        customer_id: str,
        id: str,
        feature_id: str,
        name: str,
    ) -> Entity: ...

    @overload
    def create(
        self: "Entities[AsyncHTTPClient]",
        customer_id: str,
        id: str,
        feature_id: str,
        name: str,
    ) -> Coroutine[Any, Any, Entity]: ...

    def create(
        self, customer_id: str, id: str, feature_id: str, name: str
    ) -> Union[Entity, Coroutine[Any, Any, Entity]]:
        """Create an entity for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to create the entity for.
        id: str
            The ID of the entity to create.
        feature_id: str
            The ID of the feature to create the entity for.
        name: str
            The name of the entity to create.

        Returns
        -------
        :class:`~autumn.models.entities.Entity`
            The created entity.
        """

        payload = _build_payload(
            locals(), Entities.create, ignore={"customer_id"}
        )
        return self._http.request(
            "POST", f"/customers/{customer_id}/entities", Entity, json=payload
        )

    @overload
    def delete(
        self: "Entities[HTTPClient]",
        customer_id: str,
        entity_id: str,
    ) -> Empty: ...

    @overload
    def delete(
        self: "Entities[AsyncHTTPClient]",
        customer_id: str,
        entity_id: str,
    ) -> Coroutine[Any, Any, Empty]: ...

    def delete(
        self, customer_id: str, entity_id: str
    ) -> Union[Empty, Coroutine[Any, Any, Empty]]:
        """Delete an entity for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to delete the entity for.
        entity_id: str
            The ID of the entity to delete.

        Returns
        -------
        :class:`~autumn.models.meta.Empty`
            This is a placeholder type. Treat it as :class:`None`.
        """

        return self._http.request(
            "DELETE", f"/customers/{customer_id}/entities/{entity_id}", Empty
        )

    @overload
    def transfer(
        self: "Entities[HTTPClient]",
        customer_id: str,
        to_entity_id: str,
        product_id: str,
        from_entity_id: Optional[str] = None,
    ) -> TransferProductResponse: ...

    @overload
    def transfer(
        self: "Entities[AsyncHTTPClient]",
        customer_id: str,
        to_entity_id: str,
        product_id: str,
        from_entity_id: Optional[str] = None,
    ) -> Coroutine[Any, Any, TransferProductResponse]: ...

    def transfer(
        self,
        customer_id: str,
        to_entity_id: str,
        product_id: str,
        from_entity_id: Optional[str] = None,
    ) -> Union[
        TransferProductResponse, Coroutine[Any, Any, TransferProductResponse]
    ]:
        """Transfer a product from one entity to another.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer who owns the entities.
        to_entity_id: str
            The ID of the entity to transfer the product to.
        product_id: str
            The ID of the product to transfer.
        from_entity_id: Optional[str]
            The ID of the entity to transfer the product from. If not provided,
            transfers from the customer's general balance.

        Returns
        -------
        :class:`~autumn.models.entities.TransferProductResult`
            The result of the transfer operation.
        """

        payload = _build_payload(
            locals(), Entities.transfer, ignore={"customer_id"}
        )
        return self._http.request(
            "POST",
            f"/customers/{customer_id}/transfer",
            TransferProductResponse,
            json=payload,
        )



from typing import Dict, List, Optional

from pydantic import BaseModel

from .customers import CustomerFeature, CustomerInvoice, CustomerProduct
from .env import AppEnv


class Entity(BaseModel):
    id: str
    name: str
    customer_id: str
    created_at: int
    env: AppEnv
    products: List[CustomerProduct]
    features: Dict[str, CustomerFeature]
    invoices: Optional[List[CustomerInvoice]] = None



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from pydantic import BaseModel, ConfigDict

from .products import ProductItem

if TYPE_CHECKING:
    from typing import Any, Dict

    Metadata = Dict[str, Any]

__all__ = ("Empty", "FeatureOptions", "ProductOptions", "CustomerData")


class Empty(BaseModel): ...


class FeatureOptions(BaseModel):
    feature_id: str
    quantity: int


class ProductOptions(BaseModel):
    product_id: str
    quantity: Optional[int] = None


class CustomerData(BaseModel):
    name: Optional[str] = None
    email: Optional[str] = None
    fingerprint: Optional[str] = None


class AttachProductOptions(BaseModel):
    product_id: str
    quantity: Optional[int] = None


class CheckoutLine(BaseModel):
    description: str
    amount: int
    item: ProductItem


class Cycle(BaseModel):
    starts_at: int
    total: float


class QueryDP(BaseModel):
    period: int

    model_config = ConfigDict(extra="allow")

    def get_usage(self, key: str) -> Optional[int]:
        if self.__pydantic_extra__ is None:
            return None

        return self.__pydantic_extra__.get(key)



import asyncio
from typing import Optional, Type, TypeVar

from pydantic import BaseModel

from ..error import AutumnError, AutumnHTTPError
from ..http import HTTPClient, _RetryRequestError
from ..utils import ExponentialBackoff, _build_model, _check_response

try:
    import aiohttp
except ImportError:
    raise AutumnError(
        "aiohttp is not installed. Please install it with `pip install aiohttp`",
        "missing_dependency",
    )

T = TypeVar("T", bound=BaseModel)


__all__ = ("AsyncHTTPClient",)


class AsyncHTTPClient:
    def __init__(
        self,
        base_url: str,
        version: str,
        token: str,
        attempts: int,
        *,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        self.base_url = base_url
        self.version = version
        self.session = session  # type: ignore
        self._headers = HTTPClient._build_headers(token)
        self.attempts = attempts

        self._build_url = HTTPClient._build_url

    async def request(
        self, method: str, path: str, type_: Type[T], **kwargs
    ) -> T:
        if self.session is None:
            self.session = aiohttp.ClientSession()

        url = self._build_url(self.base_url, self.version, path)

        max_attempts = self.attempts
        backoff = ExponentialBackoff()
        for attempt in range(max_attempts):
            try:
                async with self.session.request(
                    method, url, headers=self._headers, **kwargs
                ) as resp:
                    if 500 <= resp.status <= 504:
                        raise _RetryRequestError()

                    data = await resp.json()

            except (_RetryRequestError, OSError, asyncio.TimeoutError):
                if attempt == max_attempts - 1:
                    raise

                await asyncio.sleep(backoff.bedtime)
                backoff.tick()
            else:
                _check_response(resp.status, data)
                return _build_model(type_, data)

        # We should never get here. This is to appease type checkers.
        msg = f"Max retries reached for {method} {path}"
        raise AutumnHTTPError(msg, "max_retries_reached", 500)

    async def close(self):
        if self.session is not None:
            await self.session.close()
